# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax


def all(x, /, *, axis=None, keepdims=False):
  """Tests whether all input array elements evaluate to True along a specified axis."""
  return jax.numpy.all(x, axis=axis, keepdims=keepdims)


def any(x, /, *, axis=None, keepdims=False):
  """Tests whether any input array element evaluates to True along a specified axis."""
  return jax.numpy.any(x, axis=axis, keepdims=keepdims)
